import os
import sys
from importlib import import_module
from typing import List, Optional, Any
from flask import Flask, json
from yaml import load, FullLoader
from token_api.blueprints import all_blueprints

import token_api as app_root
from token_api.extensions import cors, flask_pkcs11
from token_api.tools import jsonifier
from token_api.exceptions import FormatException
from token_api.config import Config

if getattr(sys, 'frozen', False):
    APP_PATH = sys.executable
else:
    APP_PATH = app_root.__file__

APP_ROOT_FOLDER = os.path.abspath(os.path.dirname(APP_PATH))
TEMPLATE_FOLDER = os.path.join(APP_ROOT_FOLDER, 'templates')
STATIC_FOLDER = os.path.join(APP_ROOT_FOLDER, 'static')


def get_config(config_class_string: str, yaml_files: Optional[List[str]] = None) -> Config:
    """Load the Flask config from a class.
    Positional arguments:
    config_class_string -- string representation of a configuration class that will be loaded (e.g.
        'pypi_portal.config.Production').
    yaml_files -- List of YAML files to load. This is for testing, leave None in dev/production.
    Returns:
    A class object to be fed into app.config.from_object().
    """
    config_module, config_class = config_class_string.rsplit('.', 1)
    config_obj = getattr(import_module(config_module), config_class)

    # Load additional configuration settings.
    yaml_files = yaml_files or [f for f in [
        os.path.join('/', 'etc', 'token-api', 'config.yml'),
        os.path.abspath(os.path.join(APP_ROOT_FOLDER, '..', 'config.yml')),
        os.path.join(APP_ROOT_FOLDER, 'config.yml'),
    ] if os.path.exists(f)]

    if not yaml_files:
        raise FileNotFoundError('No configuration file was found!')

    additional_dict = {}
    for y in yaml_files:
        print('Loading config from {}'.format(y))
        with open(y, encoding='UTF-8') as f:
            loaded_data = load(f.read(), Loader=FullLoader)
            if isinstance(loaded_data, dict):
                additional_dict.update(loaded_data)
            else:
                raise FormatException('Failed to parse configuration {}'.format(y))

    # Merge the rest into the Flask app config.
    for key, value in additional_dict.items():
        setattr(config_obj, key, value)

    return config_obj


def create_app(config_obj: Config) -> Flask:

    """Create an application."""
    app = Flask(__name__, template_folder=TEMPLATE_FOLDER, static_folder=STATIC_FOLDER)
    config_dict = {k: getattr(config_obj, k) for k in dir(config_obj) if not k.startswith('_')}
    app.config.update(config_dict)

    for bp in all_blueprints:
        import_module(bp.import_name)
        app.register_blueprint(bp)

    class FlaskJSONProvider(json.provider.DefaultJSONProvider):
        """Custom JSONProvider which adds connexion defaults on top of Flask's"""

        @jsonifier.wrap_default
        def default(self, o: Any) -> Any:
            return super().default(o)  # type: ignore[call-arg]

    app.json = FlaskJSONProvider(app)

    cors.init_app(app)
    flask_pkcs11.init_app(app)

    with app.app_context():
        import_module('token_api.middleware')

    return app
