import os
import base64
from typing import Iterator, Optional, Dict, Union, Any
from contextlib import contextmanager
import urllib.parse
import pkcs11
from flask import Flask
from asn1crypto.x509 import Certificate


class UnknownObjectException(Exception):
    pass


def build_pkcs11_url(parameters: dict) -> str:
    pairs = []
    for key, value in parameters.items():
        if isinstance(value, bytes):
            escaped_value = '%{}'.format(value.hex('%')).upper()
        elif isinstance(value, (int, float)):
            escaped_value = urllib.parse.quote(str(value))
        else:
            escaped_value = urllib.parse.quote(value)
        pairs.append('{}={}'.format(key, escaped_value))

    return 'pkcs11:{}'.format(';'.join(pairs))


def parse_pkcs11_url(url: str) -> dict:
    if not url.startswith('pkcs11:'):
        raise AttributeError('Invalid url protocol')

    clean = url.replace('pkcs11:', '')
    pairs = clean.split(';')
    data = {}
    for pair in pairs:
        key, value = pair.split('=')
        unescaped_value: Union[str, bytes]
        if value.startswith('%'):
            unescaped_value = bytes.fromhex(value.replace('%', ''))
        else:
            unescaped_value = urllib.parse.unquote(value)
        data[key] = unescaped_value

    return data


def has_required_eidas_fields(certificate: Certificate, public_key: pkcs11.PublicKey) -> bool:
    # Tests, True=fail
    tests = [
        certificate.self_issued,  # Certificate should be not self_issued
        certificate.self_signed != 'no',  # Certificate should be not self_signed
        'key_usage' not in certificate.critical_extensions,  # key usage has to be critical section
        'digital_signature' not in certificate.key_usage_value.native,  # Key usage has to support digital_signature
        'rsa' != certificate.public_key.algorithm,  # Algorithm should be RSA
        'sha256' != certificate.hash_algo,  # Hashing algorithm should be sha256
        certificate.public_key.bit_size < 2048,  # Bit size has to be 2048 or better
        certificate.public_key.sha1 != certificate.key_identifier,  # Key identifier in cert has to match public key SHA1
        not public_key[pkcs11.Attribute.LOCAL]  # PubKey has to be locally generated on device, that means PrivKey was generated locally too
    ]

    subject_required_fields = [
        'common_name',
        'serial_number',
        'country_name'
    ]

    issuer_required_fields = [
        'organization_identifier',
        'common_name',
        'country_name',
        'organization_name',
    ]

    for test in tests:
        if test:
            return False

    for subject_required_field in subject_required_fields:
        if not certificate.subject.native.get(subject_required_field):
            return False

    for issuer_required_field in issuer_required_fields:
        if not certificate.issuer.native.get(issuer_required_field):
            return False

    return True


def inject_extended_certificate_info(obj: pkcs11.Object, public_key: Optional[pkcs11.PublicKey] = None) -> pkcs11.Object:
    if not isinstance(obj, pkcs11.Certificate):
        return obj
    certificate_der = obj[pkcs11.Attribute.VALUE]
    certificate_parsed = Certificate.load(certificate_der)

    obj.is_qualified_certificate = has_required_eidas_fields(certificate_parsed, public_key) if public_key else False
    obj.certificate_parsed = certificate_parsed
    obj.subject = dict(certificate_parsed.subject.native)
    obj.issuer = dict(certificate_parsed.issuer.native)
    obj.not_valid_after = certificate_parsed.not_valid_after
    obj.not_valid_before = certificate_parsed.not_valid_before
    obj.value = base64.b64encode(certificate_der).decode('ASCII')
    return obj


def iter_inject_extended_certificate_info(objects: Iterator[pkcs11.Object], public_keys: dict) -> Iterator[pkcs11.Object]:
    for obj in objects:
        found_public_key = public_keys.get(obj[pkcs11.Attribute.ID])
        yield inject_extended_certificate_info(obj, found_public_key)


class FlaskPkcs11:
    object_class_to_string = {
        pkcs11.ObjectClass.DATA: 'data',
        pkcs11.ObjectClass.CERTIFICATE: 'cert',
        pkcs11.ObjectClass.PUBLIC_KEY: 'public',
        pkcs11.ObjectClass.PRIVATE_KEY: 'private',
        pkcs11.ObjectClass.SECRET_KEY: 'secret',
        pkcs11.ObjectClass.HW_FEATURE: 'feature',
        pkcs11.ObjectClass.DOMAIN_PARAMETERS: 'domain'
    }

    lib: Optional[pkcs11.lib] = None
    library_path: Optional[str] = None

    def __init__(self, app: Optional[Flask] = None, **kwargs: Any):
        self._options = kwargs
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        self.library_path = app.config.get('PKCS11_LIBRARY')

    def init_library(self) -> None:
        if not self.lib and self.library_path:
            if not os.path.isfile(self.library_path):
                raise FileNotFoundError('Library {} was not found!'.format(self.library_path))

            self.lib = pkcs11.lib(self.library_path)

    def get_slots(self, *args: Any, **kwargs: Any) -> Iterator[pkcs11.Slot]:
        self.init_library()
        return self.lib.get_slots(*args, **kwargs)  # type: ignore

    def get_slot(self, slot_id: int) -> Optional[pkcs11.Slot]:
        self.init_library()
        for slot in self.lib.get_slots(token_present=True):  # type: ignore
            if slot.slot_id == slot_id:
                return slot

        return None

    def get_token(self, *args: Any, **kwargs: Any) -> pkcs11.Token:
        self.init_library()
        return self.lib.get_token(*args, **kwargs)  # type: ignore

    def find_pub_keys(self, session: pkcs11.Session) -> Dict[bytes, pkcs11.PublicKey]:
        # Find pub keys so we can match them to certs
        attrs = {
            pkcs11.constants.Attribute.CLASS: pkcs11.constants.ObjectClass.PUBLIC_KEY,
        }
        public_keys = {}
        for pub_key in session.get_objects(attrs):
            public_keys[pub_key[pkcs11.Attribute.ID]] = pub_key

        return public_keys

    def pkcs11_object_to_uri(self, pkcs11_object: Any) -> str:
        url_parameters = {}

        if isinstance(pkcs11_object, pkcs11.Slot):
            url_parameters['slot-description'] = pkcs11_object.slot_description
            url_parameters['slot-manufacturer'] = pkcs11_object.manufacturer_id
            url_parameters['slot-id'] = pkcs11_object.slot_id
        elif isinstance(pkcs11_object, pkcs11.Token):
            url_parameters['model'] = pkcs11_object.model
            url_parameters['manufacturer'] = pkcs11_object.manufacturer_id
            url_parameters['serial'] = pkcs11_object.serial.decode('UTF-8')
            url_parameters['token'] = pkcs11_object.label
        elif isinstance(pkcs11_object, pkcs11.Object):
            url_parameters['model'] = pkcs11_object.session.token.model
            url_parameters['manufacturer'] = pkcs11_object.session.token.manufacturer_id
            url_parameters['serial'] = pkcs11_object.session.token.serial.decode('UTF-8')
            url_parameters['token'] = pkcs11_object.session.token.label
            url_parameters['id'] = pkcs11_object[pkcs11.Attribute.ID]

            if pkcs11_object[pkcs11.Attribute.LABEL]:
                url_parameters['object'] = pkcs11_object[pkcs11.Attribute.LABEL]

            url_parameters['type'] = self.object_class_to_string.get(pkcs11_object[pkcs11.Attribute.CLASS], '???')

        return build_pkcs11_url(url_parameters)

    def pkcs11_object_to_dict(self, pkcs11_object: Any, ignore_related: bool = False) -> dict:
        data: Dict[str, Any] = {
            'uri': self.pkcs11_object_to_uri(pkcs11_object),
        }

        if isinstance(pkcs11_object, pkcs11.Slot):
            data['slot_description'] = pkcs11_object.slot_description
            data['slot_manufacturer'] = pkcs11_object.manufacturer_id
            data['hardware_version'] = pkcs11_object.hardware_version
            data['firmware_version'] = pkcs11_object.firmware_version
            data['slot_id'] = pkcs11_object.slot_id

            try:
                token = pkcs11_object.get_token()
                data['token'] = self.pkcs11_object_to_dict(token) if not ignore_related else None
            except pkcs11.exceptions.PKCS11Error:
                data['token'] = None

        elif isinstance(pkcs11_object, pkcs11.Token):
            data['model'] = pkcs11_object.model
            data['hardware_version'] = pkcs11_object.hardware_version
            data['firmware_version'] = pkcs11_object.firmware_version
            data['manufacturer'] = pkcs11_object.manufacturer_id
            data['serial'] = pkcs11_object.serial.decode('UTF-8')
            data['token'] = pkcs11_object.label
            data['label'] = pkcs11_object.label  # !FIXME Compatibility
            # identifier
            try:
                data['slot'] = self.pkcs11_object_to_dict(pkcs11_object.slot, ignore_related=True) if not ignore_related else None
            except pkcs11.exceptions.PKCS11Error:
                data['slot'] = None
        elif isinstance(pkcs11_object, pkcs11.Object):
            data['model'] = pkcs11_object.session.token.model
            data['manufacturer'] = pkcs11_object.session.token.manufacturer_id
            data['serial'] = pkcs11_object.session.token.serial.decode('UTF-8')
            data['token'] = pkcs11_object.session.token.label
            data['id'] = pkcs11_object[pkcs11.Attribute.ID].hex(':')

            if pkcs11_object[pkcs11.Attribute.LABEL]:
                data['object'] = pkcs11_object[pkcs11.Attribute.LABEL]

            data['type'] = self.object_class_to_string.get(pkcs11_object[pkcs11.Attribute.CLASS], '???')

            if isinstance(pkcs11_object, pkcs11.Certificate):
                data['certificate_type'] = pkcs11_object.certificate_type.value
                data['label'] = pkcs11_object[pkcs11.Attribute.LABEL]
                data['value'] = pkcs11_object.value
                data['subject'] = pkcs11_object.subject
                data['issuer'] = pkcs11_object.issuer
                data['not_valid_after'] = pkcs11_object.not_valid_after
                data['not_valid_before'] = pkcs11_object.not_valid_before
                data['is_qualified_certificate'] = pkcs11_object.is_qualified_certificate

        return data

    @contextmanager
    def get_objects_by_uri(self, url: str, user_pin: Optional[str] = None, so_pin: Optional[str] = None) -> Iterator[pkcs11.Object]:
        parsed_uri = parse_pkcs11_url(url)

        # First we need to find token
        parameters = {}
        token_label = parsed_uri.get('token')
        if token_label:
            parameters['token_label'] = token_label

        token_serial = parsed_uri.get('serial')
        if token_serial:
            parameters['token_serial'] = token_serial.encode('UTF-8')

        session = self.get_token(**parameters).open(user_pin=user_pin, so_pin=so_pin)
        try:
            # Find pub keys so we can match them to certs
            public_keys = self.find_pub_keys(session)

            attrs = {}
            object_type = parsed_uri.get('type')
            if object_type:
                swapped = {v: k for k, v in self.object_class_to_string.items()}
                object_class = swapped.get(object_type)
                if not object_class:
                    raise UnknownObjectException('Unknown object type: {}'.format(object_type))

                attrs[pkcs11.constants.Attribute.CLASS] = object_class
            yield iter_inject_extended_certificate_info(session.get_objects(attrs), public_keys)
        finally:
            session.close()

    @contextmanager
    def get_object_by_uri(self, url: str, user_pin: Optional[str] = None, so_pin: Optional[str] = None) -> pkcs11.Object:
        parsed_uri = parse_pkcs11_url(url)

        # First we need to find token
        parameters = {}
        token_label = parsed_uri.get('token')
        if token_label:
            parameters['token_label'] = token_label

        token_serial = parsed_uri.get('serial')
        if token_serial:
            parameters['token_serial'] = token_serial.encode('UTF-8')

        session = self.get_token(**parameters).open(user_pin=user_pin, so_pin=so_pin)
        try:
            # Find pub keys so we can match them to certs
            public_keys = self.find_pub_keys(session)

            key_parameters = {}
            object_label = parsed_uri.get('object')
            if object_label:
                key_parameters['label'] = object_label

            object_type = parsed_uri.get('type')
            if object_type:
                object_class = {v: k for k, v in self.object_class_to_string.items()}.get(object_type)
                if not object_class:
                    raise UnknownObjectException('Unknown object type: {}'.format(object_type))

                key_parameters['object_class'] = object_class

            object_id = parsed_uri.get('id')
            if object_id:
                key_parameters['id'] = object_id

            obj = session.get_key(**key_parameters)

            public_key = public_keys.get(obj[pkcs11.Attribute.ID]) if isinstance(obj, pkcs11.Object) else None

            yield inject_extended_certificate_info(session.get_key(**key_parameters), public_key)
        finally:
            session.close()
