import datetime
import base64
from typing import Tuple
import flask
import pkcs11
from token_api import __version__
from token_api.blueprints import api
from token_api.extensions import flask_pkcs11
from token_api.flask_pkcs11.FlaskPkcs11 import inject_extended_certificate_info

__author__ = "Adam Schubert"


@api.route('/get-slots', methods=['GET'])
def get_slots() -> Tuple[flask.Response, int]:
    try:
        return_data = [flask_pkcs11.pkcs11_object_to_dict(slot) for slot in flask_pkcs11.get_slots(token_present=True)]
    except pkcs11.exceptions.PKCS11Error:
        return_data = []

    return flask.jsonify(return_data), 200


@api.route('/get-objects', methods=['GET'])
def get_objects() -> Tuple[flask.Response, int]:
    now = datetime.datetime.now(datetime.timezone.utc)
    uri = flask.request.args.get('uri')
    valid_only = flask.request.args.get('valid_only', 'false') == 'true'
    if not uri:
        return flask.jsonify({'message': 'URI is required'}), 400

    user_pin = flask.request.args.get('user_pin')
    so_pin = flask.request.args.get('so_pin')

    try:
        with flask_pkcs11.get_objects_by_uri(uri, user_pin=user_pin, so_pin=so_pin) as found_objects:
            data = []
            for found_object in found_objects:

                if valid_only:
                    if hasattr(found_object, 'not_valid_after') and found_object.not_valid_after < now:
                        continue

                    if hasattr(found_object, 'not_valid_before') and found_object.not_valid_before > now:
                        continue

                data.append(flask_pkcs11.pkcs11_object_to_dict(found_object))

            return flask.jsonify(data), 200
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token URI {} not found'.format(uri)}), 404
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.MultipleTokensReturned:
        return flask.jsonify({'message': 'Multiple tokens match this URI!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/get-object', methods=['GET'])
def get_object() -> Tuple[flask.Response, int]:
    uri = flask.request.args.get('uri')
    if not uri:
        return flask.jsonify({'message': 'URI is required'}), 400

    user_pin = flask.request.args.get('user_pin')
    so_pin = flask.request.args.get('so_pin')

    try:
        with flask_pkcs11.get_object_by_uri(uri, user_pin=user_pin, so_pin=so_pin) as found_object:
            data = flask_pkcs11.pkcs11_object_to_dict(found_object)
            return flask.jsonify(data), 200
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token URI {} not found'.format(uri)}), 404
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/get-certificates', methods=['GET'])
def get_certificates() -> Tuple[flask.Response, int]:
    """
    DEPRECATED
    Returns:

    """
    now = datetime.datetime.now(datetime.timezone.utc)
    valid_only = flask.request.args.get('valid_only', 'false') == 'true'

    token_label = flask.request.args.get('token_label')
    if not token_label:
        return flask.jsonify({'message': 'Token label is required'}), 400

    user_pin = flask.request.args.get('user_pin')

    try:
        token_found = flask_pkcs11.get_token(token_label=token_label)
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label {} not found'.format(token_label)}), 404

    try:
        return_data = []
        with token_found.open(user_pin=user_pin) as session:
            public_keys = flask_pkcs11.find_pub_keys(session)

            # Load certificates
            attrs = {
                pkcs11.constants.Attribute.CLASS: pkcs11.constants.ObjectClass.CERTIFICATE,
            }
            for certificate in session.get_objects(attrs):
                # Ignore certificates that have no pub keys stored on device
                found_public_key = public_keys.get(certificate[pkcs11.Attribute.ID])
                if found_public_key is None:
                    continue

                extended_certificate = inject_extended_certificate_info(certificate, found_public_key)
                if valid_only and (extended_certificate.not_valid_after < now or extended_certificate.not_valid_before > now):
                    continue

                return_data.append(flask_pkcs11.pkcs11_object_to_dict(extended_certificate))

        return flask.jsonify(return_data), 200
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/get-certificate', methods=['GET'])
def get_certificate() -> Tuple[flask.Response, int]:  # pylint: disable=too-many-return-statements
    """
    DEPRECATED
    Returns:

    """
    token_label = flask.request.args.get('token_label')
    if not token_label:
        return flask.jsonify({'message': 'Token label is required'}), 400

    certificate_label = flask.request.args.get('certificate_label')
    if not certificate_label:
        return flask.jsonify({'message': 'Certificate label is required'}), 400

    user_pin = flask.request.args.get('user_pin')

    try:
        token_found = flask_pkcs11.get_token(token_label=token_label)
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label {} not found'.format(token_label)}), 404

    try:
        with token_found.open(user_pin=user_pin) as session:
            try:
                certificate = session.get_key(label=certificate_label, object_class=pkcs11.constants.ObjectClass.CERTIFICATE)
                public_key = session.get_key(object_class=pkcs11.constants.ObjectClass.PUBLIC_KEY, id=certificate[pkcs11.constants.Attribute.ID])
                extended_certificate = inject_extended_certificate_info(certificate, public_key)
                return flask.jsonify(flask_pkcs11.pkcs11_object_to_dict(extended_certificate)), 200
            except pkcs11.exceptions.NoSuchKey:
                return flask.jsonify({'message': 'Certificate was not found'}), 404
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/sign', methods=['POST'])
def sign() -> Tuple[flask.Response, int]:  # pylint: disable=too-many-return-statements, too-many-branches, too-many-statements
    """
    @TODO Remove old code and make uri required
    Returns:

    """

    if not isinstance(flask.request.json, dict):
        return flask.jsonify({'message': 'Request has to be content-type: json'}), 400

    user_pin = flask.request.json.get('user_pin')
    if not user_pin:
        return flask.jsonify({'message': 'User pin is required'}), 400

    data = flask.request.json.get('data')
    if not data:
        return flask.jsonify({'message': 'Data is required'}), 400

    data_decoded = base64.b64decode(data)

    uri = flask.request.json.get('uri')
    if uri:
        try:
            with flask_pkcs11.get_object_by_uri(uri, user_pin=user_pin) as found_object:
                if not isinstance(found_object, pkcs11.Certificate):
                    return flask.jsonify({'message': 'Requested object is not certificate.'}), 400

                private_key = found_object.session.get_key(object_class=pkcs11.constants.ObjectClass.PRIVATE_KEY, id=found_object[pkcs11.constants.Attribute.ID])
                # passing pin is not supported on older versions, try catch that
                try:
                    q_pin = user_pin if private_key[pkcs11.Attribute.ALWAYS_AUTHENTICATE] else None
                    signature = private_key.sign(data_decoded, mechanism=pkcs11.mechanisms.Mechanism.SHA256_RSA_PKCS, pin=q_pin)
                except TypeError:
                    signature = private_key.sign(data_decoded, mechanism=pkcs11.mechanisms.Mechanism.SHA256_RSA_PKCS)

                return flask.jsonify({'signature': base64.b64encode(signature).decode('UTF-8')}), 200
        except pkcs11.exceptions.NoSuchToken:
            return flask.jsonify({'message': 'Token label not found'}), 404
        except pkcs11.exceptions.NoSuchKey:
            return flask.jsonify({'message': 'Certificate was not found'}), 404
        except pkcs11.exceptions.FunctionCancelled:
            return flask.jsonify({'message': 'Request was canceled'}), 400
        except pkcs11.exceptions.PinIncorrect:
            return flask.jsonify({'message': 'PIN is incorrect!'}), 400
        except pkcs11.exceptions.PinLocked:
            return flask.jsonify({'message': 'PIN is locked!'}), 400

    token_label = flask.request.json.get('token_label')
    if not token_label:
        return flask.jsonify({'message': 'Token label is required'}), 400

    certificate_label = flask.request.json.get('certificate_label')
    if not certificate_label:
        return flask.jsonify({'message': 'Certificate label is required'}), 400

    try:
        token_found = flask_pkcs11.get_token(token_label=token_label)
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label {} not found'.format(token_label)}), 404

    try:
        with token_found.open(user_pin=user_pin) as session:
            try:
                certificate = session.get_key(label=certificate_label, object_class=pkcs11.constants.ObjectClass.CERTIFICATE)
            except pkcs11.exceptions.NoSuchKey:
                return flask.jsonify({'message': 'Certificate was not found'}), 404
            # Generate an AES key in this session
            private_key = session.get_key(object_class=pkcs11.constants.ObjectClass.PRIVATE_KEY, id=certificate[pkcs11.constants.Attribute.ID])
            # Encrypt our data
            # mechanism=pkcs11.mechanisms.Mechanism.RSA_PKCS
            try:
                # passing pin is not supported on older versions, try catch that
                try:
                    q_pin = user_pin if private_key[pkcs11.Attribute.ALWAYS_AUTHENTICATE] else None
                    signature = private_key.sign(data, mechanism=pkcs11.mechanisms.Mechanism.SHA256_RSA_PKCS, pin=q_pin)
                except TypeError:
                    signature = private_key.sign(data, mechanism=pkcs11.mechanisms.Mechanism.SHA256_RSA_PKCS)
            except pkcs11.exceptions.FunctionCancelled:
                return flask.jsonify({'message': 'Request was canceled'}), 400

            return flask.jsonify({'signature': base64.b64encode(signature).decode('UTF-8')}), 200
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/verify', methods=['POST'])
def verify() -> Tuple[flask.Response, int]:  # pylint: disable=too-many-return-statements
    if not isinstance(flask.request.json, dict):
        return flask.jsonify({'message': 'Request has to be content-type: json'}), 400

    uri = flask.request.json.get('uri')
    if not uri:
        return flask.jsonify({'message': 'uri is required'}), 400

    user_pin = flask.request.json.get('user_pin')

    data = flask.request.json.get('data')
    if not data:
        return flask.jsonify({'message': 'data is required'}), 400

    signature = flask.request.json.get('signature')
    if not signature:
        return flask.jsonify({'message': 'signature is required'}), 400

    data_decoded = base64.b64decode(data)
    signature_decoded = base64.b64decode(signature)

    try:
        with flask_pkcs11.get_object_by_uri(uri, user_pin) as found_object:
            if not isinstance(found_object, pkcs11.VerifyMixin):
                return flask.jsonify({'message': 'Object is not VerifyMixin.'}), 400

            verified = found_object.verify(data_decoded, signature_decoded, mechanism=pkcs11.mechanisms.Mechanism.SHA256_RSA_PKCS)
            return flask.jsonify({'verified': verified}), 200
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label not found'}), 404
    except pkcs11.exceptions.NoSuchKey:
        return flask.jsonify({'message': 'Certificate was not found'}), 404
    except pkcs11.exceptions.FunctionCancelled:
        return flask.jsonify({'message': 'Request was canceled'}), 400
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/encrypt', methods=['POST'])
def encrypt() -> Tuple[flask.Response, int]:  # pylint: disable=too-many-return-statements
    if not isinstance(flask.request.json, dict):
        return flask.jsonify({'message': 'Request has to be content-type: json'}), 400

    uri = flask.request.json.get('uri')
    if not uri:
        return flask.jsonify({'message': 'uri is required'}), 400

    user_pin = flask.request.json.get('user_pin')

    data = flask.request.json.get('data')
    if not data:
        return flask.jsonify({'message': 'data is required'}), 400

    data_decoded = base64.b64decode(data)

    try:
        with flask_pkcs11.get_object_by_uri(uri, user_pin=user_pin) as found_object:
            if not isinstance(found_object, pkcs11.EncryptMixin):
                return flask.jsonify({'message': 'Requested object is not EncryptMixin.'}), 400
            print(found_object)
            # passing pin is not supported on older versions, try catch that
            try:
                q_pin = user_pin if isinstance(found_object, pkcs11.PrivateKey) and found_object[pkcs11.Attribute.ALWAYS_AUTHENTICATE] else None
                encrypted = found_object.encrypt(data_decoded, pin=q_pin)
            except TypeError:
                encrypted = found_object.encrypt(data_decoded)

            return flask.jsonify({'encrypted': base64.b64encode(encrypted).decode('UTF-8')}), 200
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label not found'}), 404
    except pkcs11.exceptions.NoSuchKey:
        return flask.jsonify({'message': 'Certificate was not found'}), 404
    except pkcs11.exceptions.FunctionCancelled:
        return flask.jsonify({'message': 'Request was canceled'}), 400
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/decrypt', methods=['POST'])
def decrypt() -> Tuple[flask.Response, int]:  # pylint: disable=too-many-return-statements
    if not isinstance(flask.request.json, dict):
        return flask.jsonify({'message': 'Request has to be content-type: json'}), 400

    uri = flask.request.json.get('uri')
    if not uri:
        return flask.jsonify({'message': 'uri is required'}), 400

    user_pin = flask.request.json.get('user_pin')

    encrypted = flask.request.json.get('encrypted')
    if not encrypted:
        return flask.jsonify({'message': 'encrypted is required'}), 400

    encrypted_decoded = base64.b64decode(encrypted)

    try:
        with flask_pkcs11.get_object_by_uri(uri, user_pin) as found_object:
            if not isinstance(found_object, pkcs11.DecryptMixin):
                return flask.jsonify({'message': 'Object is not DecryptMixin.'}), 400

            data = found_object.decrypt(encrypted_decoded)
            return flask.jsonify({'data': base64.b64encode(data).decode('UTF-8')}), 200
    except pkcs11.exceptions.NoSuchToken:
        return flask.jsonify({'message': 'Token label not found'}), 404
    except pkcs11.exceptions.NoSuchKey:
        return flask.jsonify({'message': 'Certificate was not found'}), 404
    except pkcs11.exceptions.FunctionCancelled:
        return flask.jsonify({'message': 'Request was canceled'}), 400
    except pkcs11.exceptions.PinIncorrect:
        return flask.jsonify({'message': 'PIN is incorrect!'}), 400
    except pkcs11.exceptions.PinLocked:
        return flask.jsonify({'message': 'PIN is locked!'}), 400


@api.route('/get-system-info', methods=['GET'])
def get_system_info() -> Tuple[flask.Response, int]:
    return flask.jsonify({'version': __version__}), 200
